@@ -1,4 +1,5 @@
 use {
+    itertools::Itertools,
     std::{convert::From, ops::Range},
     unicode_segmentation::UnicodeSegmentation,
 };
@@ -24,11 +25,6 @@ impl LineBuffer {
         Self::default()
     }
 
-    /// Replaces the content between [`start`..`end`] with `text`
-    pub fn replace(&mut self, range: Range<usize>, text: &str) {
-        self.lines.replace_range(range, text);
-    }
-
     /// Check to see if the line buffer is empty
     pub fn is_empty(&self) -> bool {
         self.lines.is_empty()
@@ -70,6 +66,8 @@ impl LineBuffer {
     }
 
     /// Sets the current edit position
+    /// ## Unicode safety:
+    /// Not checked, inproper use may cause panics in following operations
     pub fn set_insertion_point(&mut self, offset: usize) {
         self.insertion_point = offset;
     }
@@ -179,6 +177,87 @@ impl LineBuffer {
             .unwrap_or_else(|| self.lines.len())
     }
 
+    /// Cursor position *behind* the next WORD to the right
+    pub fn big_word_right_index(&self) -> usize {
+        let mut found_ws = false;
+
+        self.lines[self.insertion_point..]
+            .split_word_bound_indices()
+            .find(|(_, word)| {
+                found_ws = found_ws || is_whitespace_str(word);
+                found_ws && !is_whitespace_str(word)
+            })
+            .map(|(i, word)| self.insertion_point + i + word.len())
+            .unwrap_or_else(|| self.lines.len())
+    }
+
+    /// Cursor position *at end of* the next word to the right
+    pub fn word_right_end_index(&self) -> usize {
+        self.lines[self.insertion_point..]
+            .split_word_bound_indices()
+            .find_map(|(i, word)| {
+                word.grapheme_indices(true)
+                    .next_back()
+                    .map(|x| self.insertion_point + x.0 + i)
+                    .filter(|x| !is_whitespace_str(word) && *x != self.insertion_point)
+            })
+            .unwrap_or_else(|| {
+                self.lines
+                    .grapheme_indices(true)
+                    .last()
+                    .map(|x| x.0)
+                    .unwrap_or(0)
+            })
+    }
+
+    /// Cursor position *at end of* the next WORD to the right
+    pub fn big_word_right_end_index(&self) -> usize {
+        self.lines[self.insertion_point..]
+            .split_word_bound_indices()
+            .tuple_windows()
+            .find_map(|((prev_i, prev_word), (_, word))| {
+                if is_whitespace_str(word) {
+                    prev_word
+                        .grapheme_indices(true)
+                        .next_back()
+                        .map(|x| self.insertion_point + x.0 + prev_i)
+                        .filter(|x| *x != self.insertion_point)
+                } else {
+                    None
+                }
+            })
+            .unwrap_or_else(|| {
+                self.lines
+                    .grapheme_indices(true)
+                    .last()
+                    .map(|x| x.0)
+                    .unwrap_or(0)
+            })
+    }
+
+    /// Cursor position *in front of* the next word to the right
+    pub fn word_right_start_index(&self) -> usize {
+        self.lines[self.insertion_point..]
+            .split_word_bound_indices()
+            .find(|(i, word)| *i != 0 && !is_whitespace_str(word))
+            .map(|(i, _)| self.insertion_point + i)
+            .unwrap_or_else(|| self.lines.len())
+    }
+
+    /// Cursor position *in front of* the next WORD to the right
+    pub fn big_word_right_start_index(&self) -> usize {
+        let mut found_ws = false;
+
+        self.lines[self.insertion_point..]
+            .split_word_bound_indices()
+            .find(|(i, word)| {
+                found_ws = found_ws || *i != 0 && is_whitespace_str(word);
+                found_ws && *i != 0 && !is_whitespace_str(word)
+            })
+            .map(|(i, _)| self.insertion_point + i)
+            .unwrap_or_else(|| self.lines.len())
+    }
+
     /// Cursor position *in front of* the next word to the left
     pub fn word_left_index(&self) -> usize {
         self.lines[..self.insertion_point]
@@ -189,6 +268,30 @@ impl LineBuffer {
             .unwrap_or(0)
     }
 
+    /// Cursor position *in front of* the next WORD to the left
+    pub fn big_word_left_index(&self) -> usize {
+        self.lines[..self.insertion_point]
+            .split_word_bound_indices()
+            .fold(None, |last_word_index, (i, word)| {
+                match (last_word_index, is_whitespace_str(word)) {
+                    (None, true) => None,
+                    (None, false) => Some(i),
+                    (Some(_), true) => None,
+                    (Some(v), false) => Some(v),
+                }
+            })
+            .unwrap_or(0)
+    }
+
+    /// Cursor position on the next whitespace
+    pub fn next_whitespace(&self) -> usize {
+        self.lines[self.insertion_point..]
+            .split_word_bound_indices()
+            .find(|(i, word)| *i != 0 && is_whitespace_str(word))
+            .map(|(i, _)| self.insertion_point + i)
+            .unwrap_or_else(|| self.lines.len())
+    }
+
     /// Move cursor position *behind* the next unicode grapheme to the right
     pub fn move_right(&mut self) {
         self.insertion_point = self.grapheme_right_index();
@@ -204,11 +307,36 @@ impl LineBuffer {
         self.insertion_point = self.word_left_index();
     }
 
+    /// Move cursor position *in front of* the next WORD to the left
+    pub fn move_big_word_left(&mut self) {
+        self.insertion_point = self.big_word_left_index();
+    }
+
     /// Move cursor position *behind* the next word to the right
     pub fn move_word_right(&mut self) {
         self.insertion_point = self.word_right_index();
     }
 
+    /// Move cursor position to the start of the next word
+    pub fn move_word_right_start(&mut self) {
+        self.insertion_point = self.word_right_start_index();
+    }
+
+    /// Move cursor position to the start of the next WORD
+    pub fn move_big_word_right_start(&mut self) {
+        self.insertion_point = self.big_word_right_start_index();
+    }
+
+    /// Move cursor position to the end of the next word
+    pub fn move_word_right_end(&mut self) {
+        self.insertion_point = self.word_right_end_index();
+    }
+
+    /// Move cursor position to the end of the next WORD
+    pub fn move_big_word_right_end(&mut self) {
+        self.insertion_point = self.big_word_right_end_index();
+    }
+
     ///Insert a single character at the insertion point and move right
     pub fn insert_char(&mut self, c: char) {
         self.lines.insert(self.insertion_point, c);
@@ -226,6 +354,17 @@ impl LineBuffer {
         self.insertion_point = self.insertion_point() + string.len();
     }
 
+    /// Inserts the system specific new line character
+    ///
+    /// - On Unix systems LF (`"\n"`)
+    /// - On Windows CRLF (`"\r\n"`)
+    pub fn insert_newline(&mut self) {
+        #[cfg(target_os = "windows")]
+        self.insert_str("\r\n");
+        #[cfg(not(target_os = "windows"))]
+        self.insert_char('\n');
+    }
+
     /// Empty buffer and reset cursor
     pub fn clear(&mut self) {
         self.lines = String::new();
@@ -264,7 +403,7 @@ impl LineBuffer {
     /// Substitute text covered by `range` in the current line
     ///
     /// Safety: Does not change the insertion point/offset and is thus not unicode safe!
-    pub(crate) fn replace_range<R>(&mut self, range: R, replace_with: &str)
+    pub fn replace_range<R>(&mut self, range: R, replace_with: &str)
     where
         R: std::ops::RangeBounds<usize>,
     {
@@ -280,6 +419,16 @@ impl LineBuffer {
             .unwrap_or(false)
     }
 
+    /// Get the grapheme immediately to the right of the cursor, if any
+    pub fn grapheme_right(&self) -> &str {
+        &self.lines[self.insertion_point..self.grapheme_right_index()]
+    }
+
+    /// Get the grapheme immediately to the left of the cursor, if any
+    pub fn grapheme_left(&self) -> &str {
+        &self.lines[self.grapheme_left_index()..self.insertion_point]
+    }
+
     /// Gets the range of the word the current edit position is pointing to
     pub fn current_word_range(&self) -> Range<usize> {
         let right_index = self.word_right_index();
@@ -325,9 +474,26 @@ impl LineBuffer {
         self.move_word_right();
     }
 
-    /// Counts the number of words in the buffer
-    pub fn word_count(&self) -> usize {
-        self.lines.trim().split_whitespace().count()
+    /// Switches the ASCII case of the current char
+    pub fn switchcase_char(&mut self) {
+        let insertion_offset = self.insertion_point();
+        let right_index = self.grapheme_right_index();
+
+        if right_index > insertion_offset {
+            let change_range = insertion_offset..right_index;
+            let swapped = self.get_buffer()[change_range.clone()]
+                .chars()
+                .map(|c| {
+                    if c.is_ascii_uppercase() {
+                        c.to_ascii_lowercase()
+                    } else {
+                        c.to_ascii_uppercase()
+                    }
+                })
+                .collect::<String>();
+            self.replace_range(change_range, &swapped);
+            self.move_right();
+        }
     }
 
     /// Capitalize the character at insertion point (or the first character
@@ -709,13 +875,31 @@ mod test {
     }
 
     #[rstest]
-    #[case("This is a te", 4)]
-    #[case("This is a test", 4)]
-    #[case("This      is a test", 4)]
-    fn word_count_works(#[case] input: &str, #[case] expected_count: usize) {
-        let line_buffer = buffer_with(input);
+    #[case("", 0, 0)] // Basecase
+    #[case("word", 0, 3)] // Cursor on top of the last grapheme of the word
+    #[case("word and another one", 0, 3)]
+    #[case("word and another one", 3, 7)] // repeat calling will move
+    #[case("word and another one", 4, 7)] // Starting from whitespace works
+    #[case("word\nline two", 0, 3)] // Multiline...
+    #[case("word\nline two", 3, 8)] // ... contineus to next word end
+    #[case("weirdö characters", 0, 5)] // Multibyte unicode at the word end (latin UTF-8 should be two bytes long)
+    #[case("weirdö characters", 5, 17)] // continue with unicode (latin UTF-8 should be two bytes long)
+    #[case("weirdö", 0, 5)] // Multibyte unicode at the buffer end is fine as well
+    #[case("weirdö", 5, 5)] // Multibyte unicode at the buffer end is fine as well
+    #[case("word😇 with emoji", 0, 3)] // (Emojis are a separate word)
+    #[case("word😇 with emoji", 3, 4)] // Moves to end of "emoji word" as it is one grapheme, on top of the first byte
+    #[case("😇", 0, 0)] // More UTF-8 shenanigans
+    fn test_move_word_right_end(
+        #[case] input: &str,
+        #[case] in_location: usize,
+        #[case] expected: usize,
+    ) {
+        let mut line_buffer = buffer_with(input);
+        line_buffer.set_insertion_point(in_location);
 
-        assert_eq!(expected_count, line_buffer.word_count());
+        line_buffer.move_word_right_end();
+
+        assert_eq!(line_buffer.insertion_point(), expected);
         line_buffer.assert_valid();
     }
 
@@ -786,6 +970,32 @@ mod test {
         line_buffer.assert_valid();
     }
 
+    #[rstest]
+    #[case("", 0, "", 0)]
+    #[case("a test", 2, "a Test", 3)]
+    #[case("a Test", 2, "a test", 3)]
+    #[case("test", 0, "Test", 1)]
+    #[case("Test", 0, "test", 1)]
+    #[case("test", 3, "tesT", 4)]
+    #[case("tesT", 3, "test", 4)]
+    #[case("ß", 0, "ß", 2)]
+    fn switchcase_char(
+        #[case] input: &str,
+        #[case] in_location: usize,
+        #[case] output: &str,
+        #[case] out_location: usize,
+    ) {
+        let mut line_buffer = buffer_with(input);
+        line_buffer.set_insertion_point(in_location);
+        line_buffer.switchcase_char();
+
+        let mut expected = buffer_with(output);
+        expected.set_insertion_point(out_location);
+
+        assert_eq!(expected, line_buffer);
+        line_buffer.assert_valid();
+    }
+
     #[rstest]
     #[case("This is a test", 13, "This is a tets", 14)]
     #[case("This is a test", 14, "This is a tets", 14)] // NOTE: Swaping works in opposite direction at last index
@@ -1239,4 +1449,124 @@ mod test {
         assert_eq!(expected, line_buffer);
         line_buffer.assert_valid();
     }
+
+    #[rstest]
+    #[case("abc def ghi", 10, 8)]
+    #[case("abc def-ghi", 10, 8)]
+    #[case("abc def.ghi", 10, 4)]
+    fn test_word_left_index(#[case] input: &str, #[case] position: usize, #[case] expected: usize) {
+        let mut line_buffer = buffer_with(input);
+        line_buffer.set_insertion_point(position);
+
+        let index = line_buffer.word_left_index();
+
+        assert_eq!(index, expected);
+    }
+
+    #[rstest]
+    #[case("abc def ghi", 10, 8)]
+    #[case("abc def-ghi", 10, 4)]
+    #[case("abc def.ghi", 10, 4)]
+    fn test_big_word_left_index(
+        #[case] input: &str,
+        #[case] position: usize,
+        #[case] expected: usize,
+    ) {
+        let mut line_buffer = buffer_with(input);
+        line_buffer.set_insertion_point(position);
+
+        let index = line_buffer.big_word_left_index();
+
+        assert_eq!(index, expected,);
+    }
+
+    #[rstest]
+    #[case("abc def ghi", 0, 4)]
+    #[case("abc-def ghi", 0, 3)]
+    #[case("abc.def ghi", 0, 8)]
+    fn test_word_right_start_index(
+        #[case] input: &str,
+        #[case] position: usize,
+        #[case] expected: usize,
+    ) {
+        let mut line_buffer = buffer_with(input);
+        line_buffer.set_insertion_point(position);
+
+        let index = line_buffer.word_right_start_index();
+
+        assert_eq!(index, expected);
+    }
+
+    #[rstest]
+    #[case("abc def ghi", 0, 4)]
+    #[case("abc-def ghi", 0, 8)]
+    #[case("abc.def ghi", 0, 8)]
+    fn test_big_word_right_start_index(
+        #[case] input: &str,
+        #[case] position: usize,
+        #[case] expected: usize,
+    ) {
+        let mut line_buffer = buffer_with(input);
+        line_buffer.set_insertion_point(position);
+
+        let index = line_buffer.big_word_right_start_index();
+
+        assert_eq!(index, expected);
+    }
+
+    #[rstest]
+    #[case("abc def ghi", 0, 2)]
+    #[case("abc-def ghi", 0, 2)]
+    #[case("abc.def ghi", 0, 6)]
+    #[case("abc", 1, 2)]
+    #[case("abc", 2, 2)]
+    #[case("abc def", 2, 6)]
+    fn test_word_right_end_index(
+        #[case] input: &str,
+        #[case] position: usize,
+        #[case] expected: usize,
+    ) {
+        let mut line_buffer = buffer_with(input);
+        line_buffer.set_insertion_point(position);
+
+        let index = line_buffer.word_right_end_index();
+
+        assert_eq!(index, expected);
+    }
+
+    #[rstest]
+    #[case("abc def ghi", 0, 2)]
+    #[case("abc-def ghi", 0, 6)]
+    #[case("abc-def ghi", 5, 6)]
+    #[case("abc-def ghi", 6, 10)]
+    #[case("abc.def ghi", 0, 6)]
+    #[case("abc", 1, 2)]
+    #[case("abc", 2, 2)]
+    #[case("abc def", 2, 6)]
+    #[case("abc-def", 6, 6)]
+    fn test_big_word_right_end_index(
+        #[case] input: &str,
+        #[case] position: usize,
+        #[case] expected: usize,
+    ) {
+        let mut line_buffer = buffer_with(input);
+        line_buffer.set_insertion_point(position);
+
+        let index = line_buffer.big_word_right_end_index();
+
+        assert_eq!(index, expected);
+    }
+
+    #[rstest]
+    #[case("abc def", 0, 3)]
+    #[case("abc def ghi", 3, 7)]
+    #[case("abc", 1, 3)]
+    fn test_next_whitespace(#[case] input: &str, #[case] position: usize, #[case] expected: usize) {
+        let mut line_buffer = buffer_with(input);
+        line_buffer.set_insertion_point(position);
+
+        let index = line_buffer.next_whitespace();
+
+        assert_eq!(index, expected);
+    }
 }
