@@ -1,30 +1,43 @@
+#[cfg(feature = "bashisms")]
+use crate::{
+    history::SearchFilter,
+    menu_functions::{parse_selection_char, ParseAction},
+};
+#[cfg(feature = "external_printer")]
+use {
+    crate::external_printer::ExternalPrinter,
+    crossbeam::channel::TryRecvError,
+    std::io::{Error, ErrorKind},
+};
 use {
     crate::{
-        completion::{CircularCompletionHandler, Completer, DefaultCompleter},
+        completion::{Completer, DefaultCompleter},
         core_editor::Editor,
         edit_mode::{EditMode, Emacs},
         enums::{EventStatus, ReedlineEvent},
         highlighter::SimpleMatchHighlighter,
         hinter::Hinter,
-        history::{FileBackedHistory, History, HistoryNavigationQuery},
-        menu::{Menu, MenuEvent, ReedlineMenu},
+        history::{
+            FileBackedHistory, History, HistoryCursor, HistoryItem, HistoryItemId,
+            HistoryNavigationQuery, HistorySessionId, SearchDirection, SearchQuery,
+        },
         painting::{Painter, PromptLines},
         prompt::{PromptEditMode, PromptHistorySearchStatus},
+        result::{ReedlineError, ReedlineErrorVariants},
         utils::text_manipulation,
-        EditCommand, ExampleHighlighter, Highlighter, Prompt, PromptHistorySearch, Signal,
-        ValidationResult, Validator,
+        EditCommand, ExampleHighlighter, Highlighter, LineBuffer, Menu, MenuEvent, Prompt,
+        PromptHistorySearch, ReedlineMenu, Signal, UndoBehavior, ValidationResult, Validator,
     },
     crossterm::{
         event,
         event::{Event, KeyCode, KeyEvent, KeyModifiers},
         terminal, Result,
     },
-    std::{borrow::Borrow, io, time::Duration},
+    std::{
+        borrow::Borrow, fs::File, io, io::Write, process::Command, time::Duration, time::SystemTime,
+    },
 };
 
-#[cfg(feature = "bashisms")]
-use crate::menu_functions::{parse_selection_char, ParseAction};
-
 // The POLL_WAIT is used to specify for how long the POLL should wait for
 // events, to accelerate the handling of paste or compound resize events. Having
 // a POLL_WAIT of zero means that every single event is treated as soon as it
@@ -79,6 +92,10 @@ pub struct Reedline {
 
     // History
     history: Box<dyn History>,
+    history_cursor: HistoryCursor,
+    history_session_id: Option<HistorySessionId>,
+    // none if history doesn't support this
+    history_last_run_id: Option<HistoryItemId>,
     input_mode: InputMode,
 
     // Validator
@@ -95,9 +112,6 @@ pub struct Reedline {
     quick_completions: bool,
     partial_completions: bool,
 
-    // Performs bash style circular rotation through the available completions
-    circular_completion_handler: CircularCompletionHandler,
-
     // Highlight the edit buffer
     highlighter: Box<dyn Highlighter>,
 
@@ -105,14 +119,22 @@ pub struct Reedline {
     hinter: Option<Box<dyn Hinter>>,
     hide_hints: bool,
 
-    // Is Some(n) read_line() should repaint prompt every `n` milliseconds
-    animate: bool,
-
     // Use ansi coloring or not
     use_ansi_coloring: bool,
 
     // Engine Menus
     menus: Vec<ReedlineMenu>,
+
+    // Text editor used to open the line buffer for editing
+    buffer_editor: Option<BufferEditor>,
+
+    #[cfg(feature = "external_printer")]
+    external_printer: Option<ExternalPrinter<String>>,
+}
+
+struct BufferEditor {
+    editor: String,
+    extension: String,
 }
 
 impl Drop for Reedline {
@@ -134,30 +156,52 @@ impl Reedline {
         let hinter = None;
         let validator = None;
         let edit_mode = Box::new(Emacs::default());
+        let hist_session_id = Self::create_history_session_id();
 
         Reedline {
             editor: Editor::default(),
             history,
+            history_cursor: HistoryCursor::new(HistoryNavigationQuery::Normal(
+                LineBuffer::default(),
+            )),
+            history_session_id: hist_session_id,
+            history_last_run_id: None,
             input_mode: InputMode::Regular,
             painter,
             edit_mode,
             completer,
             quick_completions: false,
             partial_completions: false,
-            circular_completion_handler: CircularCompletionHandler::default(),
             highlighter: buffer_highlighter,
             hinter,
             hide_hints: false,
             validator,
-            animate: false,
             use_ansi_coloring: true,
             menus: Vec::new(),
+            buffer_editor: None,
+            #[cfg(feature = "external_printer")]
+            external_printer: None,
         }
     }
 
+    /// Get a new history session id based on the current time and the first commit datetime of reedline
+    fn create_history_session_id() -> Option<HistorySessionId> {
+        let nanos = match SystemTime::now().duration_since(SystemTime::UNIX_EPOCH) {
+            Ok(n) => n.as_nanos() as i64,
+            Err(_) => 0,
+        };
+
+        Some(HistorySessionId::new(nanos))
+    }
+
+    /// Return the previously generated history session id
+    pub fn get_history_session_id(&self) -> Option<HistorySessionId> {
+        self.history_session_id
+    }
+
     /// A builder to include a [`Hinter`] in your instance of the Reedline engine
     /// # Example
-    /// ```rust,no_run
+    /// ```rust
     /// //Cargo.toml
     /// //[dependencies]
     /// //nu-ansi-term = "*"
@@ -186,7 +230,7 @@ impl Reedline {
 
     /// A builder to configure the tab completion
     /// # Example
-    /// ```rust,no_run
+    /// ```rust
     /// // Create a reedline object with tab completions support
     ///
     /// use reedline::{DefaultCompleter, Reedline};
@@ -231,17 +275,9 @@ impl Reedline {
         self
     }
 
-    /// A builder which enables or disables animations/automatic repainting of prompt.
-    /// If `repaint` is true, every second the prompt will be repainted and the clock updates
-    #[must_use]
-    pub fn with_animation(mut self, repaint: bool) -> Self {
-        self.animate = repaint;
-        self
-    }
-
     /// A builder that configures the highlighter for your instance of the Reedline engine
     /// # Example
-    /// ```rust,no_run
+    /// ```rust
     /// // Create a reedline object with highlighter support
     ///
     /// use reedline::{ExampleHighlighter, Reedline};
@@ -283,7 +319,7 @@ impl Reedline {
 
     /// A builder that configures the validator for your instance of the Reedline engine
     /// # Example
-    /// ```rust,no_run
+    /// ```rust
     /// // Create a reedline object with validator support
     ///
     /// use reedline::{DefaultValidator, Reedline};
@@ -297,6 +333,22 @@ impl Reedline {
         self
     }
 
+    /// A builder that configures the text editor used to edit the line buffer
+    /// # Example
+    /// ```rust,no_run
+    /// // Create a reedline object with vim as editor
+    ///
+    /// use reedline::{DefaultValidator, Reedline};
+    ///
+    /// let mut line_editor =
+    /// Reedline::create().with_buffer_editor("vim".into(), "nu".into());
+    /// ```
+    #[must_use]
+    pub fn with_buffer_editor(mut self, editor: String, extension: String) -> Self {
+        self.buffer_editor = Some(BufferEditor { editor, extension });
+        self
+    }
+
     /// Remove the current [`Validator`]
     #[must_use]
     pub fn disable_validator(mut self) -> Self {
@@ -334,13 +386,11 @@ impl Reedline {
     pub fn print_history(&mut self) -> Result<()> {
         let history: Vec<_> = self
             .history
-            .iter_chronologic()
-            .cloned()
-            .enumerate()
-            .collect();
+            .search(SearchQuery::everything(SearchDirection::Forward))
+            .expect("todo: error handling");
 
-        for (i, entry) in history {
-            self.print_line(&format!("{}\t{}", i, entry))?;
+        for (i, entry) in history.iter().enumerate() {
+            self.print_line(&format!("{}\t{}", i, entry.command_line))?;
         }
         Ok(())
     }
@@ -356,6 +406,29 @@ impl Reedline {
         self.history.sync()
     }
 
+    /// Check if any commands have been run.
+    ///
+    /// When no commands have been run, calling [`Self::update_last_command_context`]
+    /// does not make sense and is guaranteed to fail with a "No command run" error.
+    pub fn has_last_command_context(&self) -> bool {
+        self.history_last_run_id.is_some()
+    }
+
+    /// update the last history item with more information
+    pub fn update_last_command_context(
+        &mut self,
+        f: &dyn Fn(HistoryItem) -> HistoryItem,
+    ) -> crate::Result<()> {
+        if let Some(r) = &self.history_last_run_id {
+            self.history.update(*r, f)?;
+        } else {
+            return Err(ReedlineError(ReedlineErrorVariants::OtherHistoryError(
+                "No command run",
+            )));
+        }
+        Ok(())
+    }
+
     /// Wait for input and provide the user with a specified [`Prompt`].
     ///
     /// Returns a [`crossterm::Result`] in which the `Err` type is [`crossterm::ErrorKind`]
@@ -371,6 +444,11 @@ impl Reedline {
         result
     }
 
+    /// Returns the current contents of the input buffer.
+    pub fn current_buffer_contents(&self) -> &str {
+        self.editor.get_buffer()
+    }
+
     /// Writes `msg` to the terminal with a following carriage return and newline
     fn print_line(&mut self, msg: &str) -> Result<()> {
         self.painter.paint_line(msg)
@@ -384,6 +462,13 @@ impl Reedline {
         Ok(())
     }
 
+    /// Clear the screen and the scollback buffer of the terminal
+    pub fn clear_scrollback(&mut self) -> Result<()> {
+        self.painter.clear_scrollback()?;
+
+        Ok(())
+    }
+
     /// Helper implementing the logic for [`Reedline::read_line()`] to be wrapped
     /// in a `raw_mode` context.
     fn read_line_helper(&mut self, prompt: &dyn Prompt) -> Result<Signal> {
@@ -398,7 +483,22 @@ impl Reedline {
         loop {
             let mut paste_enter_state = false;
 
-            if event::poll(Duration::from_millis(1000))? {
+            #[cfg(feature = "external_printer")]
+            if let Some(ref external_printer) = self.external_printer {
+                // get messages from printer as crlf separated "lines"
+                let messages = Self::external_messages(external_printer)?;
+                if !messages.is_empty() {
+                    // print the message(s)
+                    self.painter.print_external_message(
+                        messages,
+                        self.editor.line_buffer(),
+                        prompt,
+                    )?;
+                    self.repaint(prompt)?;
+                }
+            }
+
+            if event::poll(Duration::from_millis(100))? {
                 let mut latest_resize = None;
 
                 // There could be multiple events queued up!
@@ -457,8 +557,6 @@ impl Reedline {
                 if let Some(ec) = last_edit_commands {
                     reedline_events.push(ReedlineEvent::Edit(ec));
                 }
-            } else if self.animate && !self.painter.exceeds_screen_size() {
-                reedline_events.push(ReedlineEvent::Repaint);
             };
 
             for event in reedline_events.drain(..) {
@@ -523,11 +621,18 @@ impl Reedline {
                 self.input_mode = InputMode::Regular;
                 Ok(EventStatus::Exits(Signal::CtrlC))
             }
-            ReedlineEvent::ClearScreen => Ok(EventStatus::Exits(Signal::CtrlL)),
+            ReedlineEvent::ClearScreen => {
+                self.painter.clear_screen()?;
+                Ok(EventStatus::Handled)
+            }
+            ReedlineEvent::ClearScrollback => {
+                self.painter.clear_scrollback()?;
+                Ok(EventStatus::Handled)
+            }
             ReedlineEvent::Enter | ReedlineEvent::HistoryHintComplete => {
-                if let Some(string) = self.history.string_at_cursor() {
-                    self.editor.set_buffer(string);
-                    self.editor.remember_undo_state(true);
+                if let Some(string) = self.history_cursor.string_at_cursor() {
+                    self.editor
+                        .set_buffer(string, UndoBehavior::CreateUndoPoint);
                 }
 
                 self.input_mode = InputMode::Regular;
@@ -544,21 +649,27 @@ impl Reedline {
             ReedlineEvent::Mouse => Ok(EventStatus::Handled),
             ReedlineEvent::Resize(width, height) => {
                 self.painter.handle_resize(width, height);
-                Ok(EventStatus::Handled)
+                Ok(EventStatus::Inapplicable)
             }
             ReedlineEvent::Repaint => {
                 // A handled Event causes a repaint
                 Ok(EventStatus::Handled)
             }
             ReedlineEvent::PreviousHistory | ReedlineEvent::Up | ReedlineEvent::SearchHistory => {
-                self.history.back();
+                self.history_cursor
+                    .back(self.history.as_ref())
+                    .expect("todo: error handling");
                 Ok(EventStatus::Handled)
             }
             ReedlineEvent::NextHistory | ReedlineEvent::Down => {
-                self.history.forward();
+                self.history_cursor
+                    .forward(self.history.as_ref())
+                    .expect("todo: error handling");
                 // Hacky way to ensure that we don't fall of into failed search going forward
-                if self.history.string_at_cursor().is_none() {
-                    self.history.back();
+                if self.history_cursor.string_at_cursor().is_none() {
+                    self.history_cursor
+                        .back(self.history.as_ref())
+                        .expect("todo: error handling");
                 }
                 Ok(EventStatus::Handled)
             }
@@ -569,10 +680,10 @@ impl Reedline {
             // TODO: Check if events should be handled
             ReedlineEvent::Right
             | ReedlineEvent::Left
-            | ReedlineEvent::ActionHandler
             | ReedlineEvent::Multiple(_)
             | ReedlineEvent::None
             | ReedlineEvent::HistoryHintWordComplete
+            | ReedlineEvent::OpenEditor
             | ReedlineEvent::Menu(_)
             | ReedlineEvent::MenuNext
             | ReedlineEvent::MenuPrevious
@@ -598,7 +709,7 @@ impl Reedline {
 
                         if self.quick_completions && menu.can_quick_complete() {
                             menu.update_values(
-                                self.editor.line_buffer(),
+                                &mut self.editor,
                                 self.completer.as_mut(),
                                 self.history.as_ref(),
                             );
@@ -611,7 +722,7 @@ impl Reedline {
                         if self.partial_completions
                             && menu.can_partially_complete(
                                 self.quick_completions,
-                                self.editor.line_buffer(),
+                                &mut self.editor,
                                 self.completer.as_mut(),
                                 self.history.as_ref(),
                             )
@@ -708,16 +819,8 @@ impl Reedline {
                 }
                 Ok(EventStatus::Inapplicable)
             }
-            ReedlineEvent::ActionHandler => {
-                let line_buffer = self.editor.line_buffer();
-                self.circular_completion_handler
-                    .handle(self.completer.as_mut(), line_buffer);
-                Ok(EventStatus::Handled)
-            }
             ReedlineEvent::Esc => {
-                self.menus
-                    .iter_mut()
-                    .for_each(|menu| menu.menu_event(MenuEvent::Deactivate));
+                self.deactivate_menus();
                 Ok(EventStatus::Handled)
             }
             ReedlineEvent::CtrlD => {
@@ -730,23 +833,25 @@ impl Reedline {
                 }
             }
             ReedlineEvent::CtrlC => {
-                self.menus
-                    .iter_mut()
-                    .for_each(|menu| menu.menu_event(MenuEvent::Deactivate));
+                self.deactivate_menus();
                 self.run_edit_commands(&[EditCommand::Clear]);
                 self.editor.reset_undo_stack();
                 Ok(EventStatus::Exits(Signal::CtrlC))
             }
             ReedlineEvent::ClearScreen => {
-                self.menus
-                    .iter_mut()
-                    .for_each(|menu| menu.menu_event(MenuEvent::Deactivate));
-                Ok(EventStatus::Exits(Signal::CtrlL))
+                self.deactivate_menus();
+                self.painter.clear_screen()?;
+                Ok(EventStatus::Handled)
+            }
+            ReedlineEvent::ClearScrollback => {
+                self.deactivate_menus();
+                self.painter.clear_scrollback()?;
+                Ok(EventStatus::Handled)
             }
             ReedlineEvent::Enter => {
                 for menu in self.menus.iter_mut() {
                     if menu.is_active() {
-                        menu.replace_in_buffer(self.editor.line_buffer());
+                        menu.replace_in_buffer(&mut self.editor);
                         menu.menu_event(MenuEvent::Deactivate);
 
                         return Ok(EventStatus::Handled);
@@ -764,18 +869,20 @@ impl Reedline {
                         self.hide_hints = true;
                         // Additional repaint to show the content without hints etc.
                         self.repaint(prompt)?;
-                        self.history.append(self.editor.get_buffer());
+                        let buf = self.editor.get_buffer();
+                        if !buf.is_empty() {
+                            let mut entry = HistoryItem::from_command_line(buf);
+                            entry.session_id = self.history_session_id;
+                            let entry = self.history.save(entry).expect("todo: error handling");
+                            self.history_last_run_id = entry.id;
+                        }
                         self.run_edit_commands(&[EditCommand::Clear]);
                         self.editor.reset_undo_stack();
 
                         Ok(EventStatus::Exits(Signal::Success(buffer)))
                     }
                     Some(ValidationResult::Incomplete) => {
-                        #[cfg(windows)]
-                        {
-                            self.run_edit_commands(&[EditCommand::InsertChar('\r')]);
-                        }
-                        self.run_edit_commands(&[EditCommand::InsertChar('\n')]);
+                        self.run_edit_commands(&[EditCommand::InsertNewline]);
 
                         Ok(EventStatus::Handled)
                     }
@@ -791,7 +898,7 @@ impl Reedline {
                     if self.quick_completions && menu.can_quick_complete() {
                         menu.menu_event(MenuEvent::Edit(self.quick_completions));
                         menu.update_values(
-                            self.editor.line_buffer(),
+                            &mut self.editor,
                             self.completer.as_mut(),
                             self.history.as_ref(),
                         );
@@ -810,9 +917,10 @@ impl Reedline {
 
                 Ok(EventStatus::Handled)
             }
+            ReedlineEvent::OpenEditor => self.open_editor().map(|_| EventStatus::Handled),
             ReedlineEvent::Resize(width, height) => {
                 self.painter.handle_resize(width, height);
-                Ok(EventStatus::Handled)
+                Ok(EventStatus::Inapplicable)
             }
             ReedlineEvent::Repaint => {
                 // A handled Event causes a repaint
@@ -843,9 +951,6 @@ impl Reedline {
                 Ok(EventStatus::Handled)
             }
             ReedlineEvent::SearchHistory => {
-                // Make sure we are able to undo the result of a reverse history search
-                self.editor.remember_undo_state(true);
-
                 self.enter_history_search();
                 Ok(EventStatus::Handled)
             }
@@ -892,39 +997,52 @@ impl Reedline {
         self.menus.iter_mut().find(|menu| menu.is_active())
     }
 
+    fn deactivate_menus(&mut self) {
+        self.menus
+            .iter_mut()
+            .for_each(|menu| menu.menu_event(MenuEvent::Deactivate));
+    }
+
     fn previous_history(&mut self) {
         if self.input_mode != InputMode::HistoryTraversal {
             self.input_mode = InputMode::HistoryTraversal;
-            self.set_history_navigation_based_on_line_buffer();
+            self.history_cursor =
+                HistoryCursor::new(self.get_history_navigation_based_on_line_buffer());
         }
 
-        self.history.back();
+        self.history_cursor
+            .back(self.history.as_ref())
+            .expect("todo: error handling");
         self.update_buffer_from_history();
-        self.editor.move_to_start();
-        self.editor.move_to_line_end();
+        self.editor.move_to_start(UndoBehavior::HistoryNavigation);
+        self.editor
+            .move_to_line_end(UndoBehavior::HistoryNavigation);
     }
 
     fn next_history(&mut self) {
         if self.input_mode != InputMode::HistoryTraversal {
             self.input_mode = InputMode::HistoryTraversal;
-            self.set_history_navigation_based_on_line_buffer();
+            self.history_cursor =
+                HistoryCursor::new(self.get_history_navigation_based_on_line_buffer());
         }
 
-        self.history.forward();
+        self.history_cursor
+            .forward(self.history.as_ref())
+            .expect("todo: error handling");
         self.update_buffer_from_history();
-        self.editor.move_to_end();
+        self.editor.move_to_end(UndoBehavior::HistoryNavigation);
     }
 
     /// Enable the search and navigation through the history from the line buffer prompt
     ///
     /// Enables either prefix search with output in the line buffer or simple traversal
-    fn set_history_navigation_based_on_line_buffer(&mut self) {
+    fn get_history_navigation_based_on_line_buffer(&self) -> HistoryNavigationQuery {
         if self.editor.is_empty() || !self.editor.is_cursor_at_buffer_end() {
             // Perform bash-style basic up/down entry walking
-            self.history.set_navigation(HistoryNavigationQuery::Normal(
+            HistoryNavigationQuery::Normal(
                 // Hack: Tight coupling point to be able to restore previously typed input
                 self.editor.line_buffer().clone(),
-            ));
+            )
         } else {
             // Prefix search like found in fish, zsh, etc.
             // Search string is set once from the current buffer
@@ -932,8 +1050,7 @@ impl Reedline {
             // Continuing with typing will leave the search
             // but next invocation of this method will start the next search
             let buffer = self.editor.get_buffer().to_string();
-            self.history
-                .set_navigation(HistoryNavigationQuery::PrefixSearch(buffer));
+            HistoryNavigationQuery::PrefixSearch(buffer)
         }
     }
 
@@ -941,9 +1058,9 @@ impl Reedline {
     ///
     /// This mode uses a separate prompt and handles keybindings slightly differently!
     fn enter_history_search(&mut self) {
+        self.history_cursor =
+            HistoryCursor::new(HistoryNavigationQuery::SubstringSearch("".to_string()));
         self.input_mode = InputMode::HistorySearch;
-        self.history
-            .set_navigation(HistoryNavigationQuery::SubstringSearch("".to_string()));
     }
 
     /// Dispatches the applicable [`EditCommand`] actions for editing the history search string.
@@ -953,30 +1070,32 @@ impl Reedline {
         for command in commands {
             match command {
                 EditCommand::InsertChar(c) => {
-                    let navigation = self.history.get_navigation();
+                    let navigation = self.history_cursor.get_navigation();
                     if let HistoryNavigationQuery::SubstringSearch(mut substring) = navigation {
                         substring.push(*c);
-                        self.history
-                            .set_navigation(HistoryNavigationQuery::SubstringSearch(substring));
+                        self.history_cursor =
+                            HistoryCursor::new(HistoryNavigationQuery::SubstringSearch(substring));
                     } else {
-                        self.history
-                            .set_navigation(HistoryNavigationQuery::SubstringSearch(String::from(
-                                *c,
-                            )));
+                        self.history_cursor = HistoryCursor::new(
+                            HistoryNavigationQuery::SubstringSearch(String::from(*c)),
+                        );
                     }
-                    self.history.back();
+                    self.history_cursor
+                        .back(self.history.as_mut())
+                        .expect("todo: error handling");
                 }
                 EditCommand::Backspace => {
-                    let navigation = self.history.get_navigation();
+                    let navigation = self.history_cursor.get_navigation();
 
                     if let HistoryNavigationQuery::SubstringSearch(substring) = navigation {
                         let new_substring = text_manipulation::remove_last_grapheme(&substring);
 
-                        self.history
-                            .set_navigation(HistoryNavigationQuery::SubstringSearch(
-                                new_substring.to_string(),
-                            ));
-                        self.history.back();
+                        self.history_cursor = HistoryCursor::new(
+                            HistoryNavigationQuery::SubstringSearch(new_substring.to_string()),
+                        );
+                        self.history_cursor
+                            .back(self.history.as_mut())
+                            .expect("todo: error handling");
                     }
                 }
                 _ => {
@@ -991,23 +1110,24 @@ impl Reedline {
     /// When using the up/down traversal or fish/zsh style prefix search update the main line buffer accordingly.
     /// Not used for the separate modal reverse search!
     fn update_buffer_from_history(&mut self) {
-        match self.history.get_navigation() {
+        match self.history_cursor.get_navigation() {
             HistoryNavigationQuery::Normal(original) => {
-                if let Some(buffer_to_paint) = self.history.string_at_cursor() {
-                    self.editor.set_buffer(buffer_to_paint.clone());
-                    self.editor.set_insertion_point(buffer_to_paint.len());
+                if let Some(buffer_to_paint) = self.history_cursor.string_at_cursor() {
+                    self.editor
+                        .set_buffer(buffer_to_paint, UndoBehavior::HistoryNavigation);
                 } else {
                     // Hack
-                    self.editor.set_line_buffer(original);
+                    self.editor
+                        .set_line_buffer(original, UndoBehavior::HistoryNavigation);
                 }
             }
             HistoryNavigationQuery::PrefixSearch(prefix) => {
-                if let Some(prefix_result) = self.history.string_at_cursor() {
-                    self.editor.set_buffer(prefix_result.clone());
-                    self.editor.set_insertion_point(prefix_result.len());
+                if let Some(prefix_result) = self.history_cursor.string_at_cursor() {
+                    self.editor
+                        .set_buffer(prefix_result, UndoBehavior::HistoryNavigation);
                 } else {
-                    self.editor.set_buffer(prefix.clone());
-                    self.editor.set_insertion_point(prefix.len());
+                    self.editor
+                        .set_buffer(prefix, UndoBehavior::HistoryNavigation);
                 }
             }
             HistoryNavigationQuery::SubstringSearch(_) => todo!(),
@@ -1015,14 +1135,15 @@ impl Reedline {
     }
 
     /// Executes [`EditCommand`] actions by modifying the internal state appropriately. Does not output itself.
-    fn run_edit_commands(&mut self, commands: &[EditCommand]) {
+    pub fn run_edit_commands(&mut self, commands: &[EditCommand]) {
         if self.input_mode == InputMode::HistoryTraversal {
             if matches!(
-                self.history.get_navigation(),
+                self.history_cursor.get_navigation(),
                 HistoryNavigationQuery::Normal(_)
             ) {
-                if let Some(string) = self.history.string_at_cursor() {
-                    self.editor.set_buffer(string);
+                if let Some(string) = self.history_cursor.string_at_cursor() {
+                    self.editor
+                        .set_buffer(string, UndoBehavior::HistoryNavigation);
                 }
             }
             self.input_mode = InputMode::Regular;
@@ -1056,7 +1177,7 @@ impl Reedline {
 
     /// Checks if hints should be displayed and are able to be completed
     fn hints_active(&self) -> bool {
-        !self.hide_hints && self.input_mode == InputMode::Regular
+        !self.hide_hints && matches!(self.input_mode, InputMode::Regular)
     }
 
     /// Repaint of either the buffer or the parts for reverse history search
@@ -1085,23 +1206,72 @@ impl Reedline {
             .index
             .zip(parsed.marker)
             .and_then(|(index, indicator)| match parsed.action {
+                ParseAction::LastCommand => self
+                    .history
+                    .search(SearchQuery {
+                        direction: SearchDirection::Backward,
+                        start_time: None,
+                        end_time: None,
+                        start_id: None,
+                        end_id: None,
+                        limit: Some(1), // fetch the latest one entries
+                        filter: SearchFilter::anything(),
+                    })
+                    .unwrap_or_else(|_| Vec::new())
+                    .get(index.saturating_sub(1))
+                    .map(|history| {
+                        (
+                            parsed.remainder.len(),
+                            indicator.len(),
+                            history.command_line.clone(),
+                        )
+                    }),
                 ParseAction::BackwardSearch => self
                     .history
-                    .iter_chronologic()
-                    .rev()
-                    .nth(index.saturating_sub(1))
-                    .map(|history| (parsed.remainder.len(), indicator.len(), history.clone())),
+                    .search(SearchQuery {
+                        direction: SearchDirection::Backward,
+                        start_time: None,
+                        end_time: None,
+                        start_id: None,
+                        end_id: None,
+                        limit: Some(index as i64), // fetch the latest n entries
+                        filter: SearchFilter::anything(),
+                    })
+                    .unwrap_or_else(|_| Vec::new())
+                    .get(index.saturating_sub(1))
+                    .map(|history| {
+                        (
+                            parsed.remainder.len(),
+                            indicator.len(),
+                            history.command_line.clone(),
+                        )
+                    }),
                 ParseAction::ForwardSearch => self
                     .history
-                    .iter_chronologic()
-                    .nth(index)
-                    .map(|history| (parsed.remainder.len(), indicator.len(), history.clone())),
+                    .search(SearchQuery {
+                        direction: SearchDirection::Forward,
+                        start_time: None,
+                        end_time: None,
+                        start_id: None,
+                        end_id: None,
+                        limit: Some((index + 1) as i64), // fetch the oldest n entries
+                        filter: SearchFilter::anything(),
+                    })
+                    .unwrap_or_else(|_| Vec::new())
+                    .get(index)
+                    .map(|history| {
+                        (
+                            parsed.remainder.len(),
+                            indicator.len(),
+                            history.command_line.clone(),
+                        )
+                    }),
                 ParseAction::LastToken => self
                     .history
-                    .iter_chronologic()
-                    .rev()
-                    .next()
-                    .and_then(|history| history.split_whitespace().rev().next())
+                    .search(SearchQuery::last_with_search(SearchFilter::anything()))
+                    .unwrap_or_else(|_| Vec::new())
+                    .get(0)
+                    .and_then(|history| history.command_line.split_whitespace().rev().next())
                     .map(|token| (parsed.remainder.len(), indicator.len(), token.to_string())),
             });
 
@@ -1117,23 +1287,54 @@ impl Reedline {
         }
     }
 
+    fn open_editor(&mut self) -> Result<()> {
+        match &self.buffer_editor {
+            None => Ok(()),
+            Some(BufferEditor { editor, extension }) => {
+                let temp_directory = std::env::temp_dir();
+                let temp_file = temp_directory.join(format!("reedline_buffer.{}", extension));
+
+                {
+                    let mut file = File::create(temp_file.clone())?;
+                    write!(file, "{}", self.editor.get_buffer())?;
+                }
+
+                {
+                    let mut process = Command::new(editor);
+                    process.arg(temp_file.as_path());
+
+                    let mut child = process.spawn()?;
+                    child.wait()?;
+                }
+
+                let res = std::fs::read_to_string(temp_file)?;
+                let res = res.trim_end().to_string();
+
+                self.editor.set_buffer(res, UndoBehavior::CreateUndoPoint);
+
+                Ok(())
+            }
+        }
+    }
+
     /// Repaint logic for the history reverse search
     ///
     /// Overwrites the prompt indicator and highlights the search string
     /// separately from the result buffer.
     fn history_search_paint(&mut self, prompt: &dyn Prompt) -> Result<()> {
-        let navigation = self.history.get_navigation();
+        let navigation = self.history_cursor.get_navigation();
 
         if let HistoryNavigationQuery::SubstringSearch(substring) = navigation {
-            let status = if !substring.is_empty() && self.history.string_at_cursor().is_none() {
-                PromptHistorySearchStatus::Failing
-            } else {
-                PromptHistorySearchStatus::Passing
-            };
+            let status =
+                if !substring.is_empty() && self.history_cursor.string_at_cursor().is_none() {
+                    PromptHistorySearchStatus::Failing
+                } else {
+                    PromptHistorySearchStatus::Passing
+                };
 
             let prompt_history_search = PromptHistorySearch::new(status, substring.clone());
 
-            let res_string = self.history.string_at_cursor().unwrap_or_default();
+            let res_string = self.history_cursor.string_at_cursor().unwrap_or_default();
 
             // Highlight matches
             let res_string = if self.use_ansi_coloring {
@@ -1205,7 +1406,7 @@ impl Reedline {
         for menu in self.menus.iter_mut() {
             if menu.is_active() {
                 menu.update_working_details(
-                    self.editor.line_buffer(),
+                    &mut self.editor,
                     self.completer.as_mut(),
                     self.history.as_ref(),
                     &self.painter,
@@ -1218,6 +1419,36 @@ impl Reedline {
         self.painter
             .repaint_buffer(prompt, &lines, menu, self.use_ansi_coloring)
     }
+
+    /// Adds an external printer
+    #[cfg(feature = "external_printer")]
+    pub fn with_external_printer(mut self, printer: ExternalPrinter<String>) -> Self {
+        self.external_printer = Some(printer);
+        self
+    }
+
+    #[cfg(feature = "external_printer")]
+    fn external_messages(external_printer: &ExternalPrinter<String>) -> Result<Vec<String>> {
+        let mut messages = Vec::new();
+        loop {
+            let result = external_printer.receiver().try_recv();
+            match result {
+                Ok(line) => {
+                    messages.push(line);
+                }
+                Err(TryRecvError::Empty) => {
+                    break;
+                }
+                Err(TryRecvError::Disconnected) => {
+                    return Err(Error::new(
+                        ErrorKind::NotConnected,
+                        TryRecvError::Disconnected,
+                    ));
+                }
+            }
+        }
+        Ok(messages)
+    }
 }
 
 #[test]
