@@ -1,6 +1,7 @@
 use super::{menu_functions::find_common_string, Menu, MenuEvent, MenuTextStyle};
 use crate::{
-    menu_functions::string_difference, painting::Painter, Completer, LineBuffer, Suggestion,
+    core_editor::Editor, menu_functions::string_difference, painting::Painter, Completer,
+    Suggestion, UndoBehavior,
 };
 use nu_ansi_term::{ansi::RESET, Style};
 
@@ -244,7 +245,7 @@ impl ColumnarMenu {
     fn move_left(&mut self) {
         self.col_pos = if let Some(row) = self.col_pos.checked_sub(1) {
             row
-        } else if self.index() == self.values.len() - 1 {
+        } else if self.index() + 1 == self.values.len() {
             0
         } else {
             self.get_cols().saturating_sub(1)
@@ -254,7 +255,7 @@ impl ColumnarMenu {
     /// Move menu cursor element
     fn move_right(&mut self) {
         let new_col = self.col_pos + 1;
-        self.col_pos = if new_col >= self.get_cols() || self.index() + 1 > self.values.len() - 1 {
+        self.col_pos = if new_col >= self.get_cols() || self.index() + 2 > self.values.len() {
             0
         } else {
             new_col
@@ -464,13 +465,13 @@ impl Menu for ColumnarMenu {
     fn can_partially_complete(
         &mut self,
         values_updated: bool,
-        line_buffer: &mut LineBuffer,
+        editor: &mut Editor,
         completer: &mut dyn Completer,
     ) -> bool {
         // If the values were already updated (e.g. quick completions are true)
         // there is no need to update the values from the menu
         if !values_updated {
-            self.update_values(line_buffer, completer);
+            self.update_values(editor, completer);
         }
 
         let values = self.get_values();
@@ -478,8 +479,12 @@ impl Menu for ColumnarMenu {
             let index = index.min(value.len());
             let matching = &value[0..index];
 
-            if !matching.is_empty() {
-                line_buffer.replace(span.start..span.end, matching);
+            // make sure that the partial completion does not overwrite user entered input
+            let extends_input = matching.starts_with(&editor.get_buffer()[span.start..span.end]);
+
+            if !matching.is_empty() && extends_input {
+                let mut line_buffer = editor.line_buffer().clone();
+                line_buffer.replace_range(span.start..span.end, matching);
 
                 let offset = if matching.len() < (span.end - span.start) {
                     line_buffer
@@ -490,10 +495,11 @@ impl Menu for ColumnarMenu {
                 };
 
                 line_buffer.set_insertion_point(offset);
+                editor.set_line_buffer(line_buffer, UndoBehavior::CreateUndoPoint);
 
                 // The values need to be updated because the spans need to be
                 // recalculated for accurate replacement in the string
-                self.update_values(line_buffer, completer);
+                self.update_values(editor, completer);
 
                 true
             } else {
@@ -519,10 +525,10 @@ impl Menu for ColumnarMenu {
     }
 
     /// Updates menu values
-    fn update_values(&mut self, line_buffer: &mut LineBuffer, completer: &mut dyn Completer) {
+    fn update_values(&mut self, editor: &mut Editor, completer: &mut dyn Completer) {
         if self.only_buffer_difference {
             if let Some(old_string) = &self.input {
-                let (start, input) = string_difference(line_buffer.get_buffer(), old_string);
+                let (start, input) = string_difference(editor.get_buffer(), old_string);
                 if !input.is_empty() {
                     self.values = completer.complete(input, start);
                     self.reset_position();
@@ -534,9 +540,8 @@ impl Menu for ColumnarMenu {
             // editing a multiline buffer.
             // Also, by replacing the new line character with a space, the insert
             // position is maintain in the line buffer.
-            let trimmed_buffer = line_buffer.get_buffer().replace('\n', " ");
-            self.values =
-                completer.complete(trimmed_buffer.as_str(), line_buffer.insertion_point());
+            let trimmed_buffer = editor.get_buffer().replace('\n', " ");
+            self.values = completer.complete(trimmed_buffer.as_str(), editor.insertion_point());
             self.reset_position();
         }
     }
@@ -545,7 +550,7 @@ impl Menu for ColumnarMenu {
     /// collected from the completer
     fn update_working_details(
         &mut self,
-        line_buffer: &mut LineBuffer,
+        editor: &mut Editor,
         completer: &mut dyn Completer,
         painter: &Painter,
     ) {
@@ -614,13 +619,13 @@ impl Menu for ColumnarMenu {
                     self.reset_position();
 
                     self.input = if self.only_buffer_difference {
-                        Some(line_buffer.get_buffer().to_string())
+                        Some(editor.get_buffer().to_string())
                     } else {
                         None
                     };
 
                     if !updated {
-                        self.update_values(line_buffer, completer);
+                        self.update_values(editor, completer);
                     }
                 }
                 MenuEvent::Deactivate => self.active = false,
@@ -628,7 +633,7 @@ impl Menu for ColumnarMenu {
                     self.reset_position();
 
                     if !updated {
-                        self.update_values(line_buffer, completer);
+                        self.update_values(editor, completer);
                     }
                 }
                 MenuEvent::NextElement => self.move_next(),
@@ -645,15 +650,26 @@ impl Menu for ColumnarMenu {
     }
 
     /// The buffer gets replaced in the Span location
-    fn replace_in_buffer(&self, line_buffer: &mut LineBuffer) {
-        if let Some(Suggestion { value, span, .. }) = self.get_value() {
-            let start = span.start.min(line_buffer.len());
-            let end = span.end.min(line_buffer.len());
-            line_buffer.replace(start..end, &value);
+    fn replace_in_buffer(&self, editor: &mut Editor) {
+        if let Some(Suggestion {
+            mut value,
+            span,
+            append_whitespace,
+            ..
+        }) = self.get_value()
+        {
+            let start = span.start.min(editor.line_buffer().len());
+            let end = span.end.min(editor.line_buffer().len());
+            if append_whitespace {
+                value.push(' ');
+            }
+            let mut line_buffer = editor.line_buffer().clone();
+            line_buffer.replace_range(start..end, &value);
 
             let mut offset = line_buffer.insertion_point();
             offset += value.len().saturating_sub(end.saturating_sub(start));
             line_buffer.set_insertion_point(offset);
+            editor.set_line_buffer(line_buffer, UndoBehavior::CreateUndoPoint);
         }
     }
 
@@ -705,3 +721,93 @@ impl Menu for ColumnarMenu {
         }
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use crate::Span;
+
+    use super::*;
+
+    macro_rules! partial_completion_tests {
+        (name: $test_group_name:ident, completions: $completions:expr, test_cases: $($name:ident: $value:expr,)*) => {
+            mod $test_group_name {
+                use crate::{menu::Menu, ColumnarMenu, core_editor::Editor, enums::UndoBehavior};
+                use super::FakeCompleter;
+
+                $(
+                    #[test]
+                    fn $name() {
+                        let (input, expected) = $value;
+                        let mut menu = ColumnarMenu::default();
+                        let mut editor = Editor::default();
+                        editor.set_buffer(input.to_string(), UndoBehavior::CreateUndoPoint);
+                        let mut completer = FakeCompleter::new(&$completions);
+
+                        menu.can_partially_complete(false, &mut editor, &mut completer);
+
+                        assert_eq!(editor.get_buffer(), expected);
+                    }
+                )*
+            }
+        }
+    }
+
+    partial_completion_tests! {
+        name: partial_completion_prefix_matches,
+        completions: ["build.rs", "build-all.sh"],
+
+        test_cases:
+            empty_completes_prefix: ("", "build"),
+            partial_completes_shared_prefix: ("bui", "build"),
+            full_prefix_completes_nothing: ("build", "build"),
+    }
+
+    partial_completion_tests! {
+        name: partial_completion_fuzzy_matches,
+        completions: ["build.rs", "build-all.sh", "prepare-build.sh"],
+
+        test_cases:
+            no_shared_prefix_completes_nothing: ("", ""),
+            shared_prefix_completes_nothing: ("bui", "bui"),
+    }
+
+    partial_completion_tests! {
+        name: partial_completion_fuzzy_same_prefix_matches,
+        completions: ["build.rs", "build-all.sh", "build-all-tests.sh"],
+
+        test_cases:
+            // assure "all" does not get replaced with shared prefix "build"
+            completes_no_shared_prefix: ("all", "all"),
+    }
+
+    struct FakeCompleter {
+        completions: Vec<String>,
+    }
+
+    impl FakeCompleter {
+        fn new(completions: &[&str]) -> Self {
+            Self {
+                completions: completions.iter().map(|c| c.to_string()).collect(),
+            }
+        }
+    }
+
+    impl Completer for FakeCompleter {
+        fn complete(&mut self, _line: &str, pos: usize) -> Vec<Suggestion> {
+            self.completions
+                .iter()
+                .map(|c| fake_suggestion(c, pos))
+                .collect()
+        }
+    }
+
+    fn fake_suggestion(name: &str, pos: usize) -> Suggestion {
+        Suggestion {
+            value: name.to_string(),
+            description: None,
+            extra: None,
+            span: Span { start: 0, end: pos },
+            append_whitespace: false,
+        }
+    }
+}
