@@ -1,3 +1,5 @@
+use crate::{core_editor::Editor, UndoBehavior};
+
 use {
     super::{
         menu_functions::{parse_selection_char, string_difference},
@@ -5,10 +7,11 @@ use {
     },
     crate::{
         painting::{estimate_single_line_wraps, Painter},
-        Completer, LineBuffer, Suggestion,
+        Completer, Suggestion,
     },
     nu_ansi_term::{ansi::RESET, Style},
     std::iter::Sum,
+    unicode_width::UnicodeWidthStr,
 };
 
 const SELECTION_CHAR: char = '!';
@@ -219,7 +222,13 @@ impl ListMenu {
                         None => (lines, None),
                         Some(total_lines) => {
                             let new_total_lines = total_lines
-                                + self.number_of_lines(&suggestion.value, painter.screen_width());
+                                + self.number_of_lines(
+                                    &suggestion.value,
+                                    //  to account for the index and the indicator e.g. 0: XXXX
+                                    painter.screen_width().saturating_sub(
+                                        self.indicator().width() as u16 + count_digits(lines),
+                                    ),
+                                );
 
                             if new_total_lines < available_lines {
                                 (lines + 1, Some(new_total_lines))
@@ -364,7 +373,7 @@ impl Menu for ListMenu {
     fn can_partially_complete(
         &mut self,
         _values_updated: bool,
-        _line_buffer: &mut LineBuffer,
+        _editor: &mut Editor,
         _completer: &mut dyn Completer,
     ) -> bool {
         false
@@ -385,7 +394,8 @@ impl Menu for ListMenu {
     }
 
     /// Collecting the value from the completer to be shown in the menu
-    fn update_values(&mut self, line_buffer: &mut LineBuffer, completer: &mut dyn Completer) {
+    fn update_values(&mut self, editor: &mut Editor, completer: &mut dyn Completer) {
+        let line_buffer = editor.line_buffer();
         let (start, input) = if self.only_buffer_difference {
             match &self.input {
                 Some(old_string) => {
@@ -456,21 +466,33 @@ impl Menu for ListMenu {
     }
 
     /// The buffer gets cleared with the actual value
-    fn replace_in_buffer(&self, line_buffer: &mut LineBuffer) {
-        if let Some(Suggestion { value, span, .. }) = self.get_value() {
-            let start = span.start.min(line_buffer.len());
-            let end = span.end.min(line_buffer.len());
-            line_buffer.replace(start..end, &value);
+    fn replace_in_buffer(&self, editor: &mut Editor) {
+        if let Some(Suggestion {
+            mut value,
+            span,
+            append_whitespace,
+            ..
+        }) = self.get_value()
+        {
+            let buffer_len = editor.line_buffer().len();
+            let start = span.start.min(buffer_len);
+            let end = span.end.min(buffer_len);
+            if append_whitespace {
+                value.push(' ');
+            }
+            let mut line_buffer = editor.line_buffer().clone();
+            line_buffer.replace_range(start..end, &value);
 
             let mut offset = line_buffer.insertion_point();
             offset += value.len().saturating_sub(end.saturating_sub(start));
             line_buffer.set_insertion_point(offset);
+            editor.set_line_buffer(line_buffer, UndoBehavior::CreateUndoPoint);
         }
     }
 
     fn update_working_details(
         &mut self,
-        line_buffer: &mut LineBuffer,
+        editor: &mut Editor,
         completer: &mut dyn Completer,
         painter: &Painter,
     ) {
@@ -480,12 +502,12 @@ impl Menu for ListMenu {
                     self.reset_position();
 
                     self.input = if self.only_buffer_difference {
-                        Some(line_buffer.get_buffer().to_string())
+                        Some(editor.get_buffer().to_string())
                     } else {
                         None
                     };
 
-                    self.update_values(line_buffer, completer);
+                    self.update_values(editor, completer);
 
                     self.pages.push(Page {
                         size: self.printable_entries(painter),
@@ -497,7 +519,7 @@ impl Menu for ListMenu {
                     self.input = None;
                 }
                 MenuEvent::Edit(_) => {
-                    self.update_values(line_buffer, completer);
+                    self.update_values(editor, completer);
                     self.pages.push(Page {
                         size: self.printable_entries(painter),
                         full: false,
@@ -509,7 +531,7 @@ impl Menu for ListMenu {
                     if let Some(page) = self.pages.get(self.page) {
                         if new_pos >= page.size as u16 {
                             self.event = Some(MenuEvent::NextPage);
-                            self.update_working_details(line_buffer, completer, painter);
+                            self.update_working_details(editor, completer, painter);
                         } else {
                             self.row_position = new_pos;
                         }
@@ -530,7 +552,7 @@ impl Menu for ListMenu {
                         }
 
                         self.event = Some(MenuEvent::PreviousPage);
-                        self.update_working_details(line_buffer, completer, painter);
+                        self.update_working_details(editor, completer, painter);
                     }
                 }
                 MenuEvent::NextPage => {
@@ -550,12 +572,12 @@ impl Menu for ListMenu {
                             }
                         }
 
-                        self.update_values(line_buffer, completer);
+                        self.update_values(editor, completer);
                         self.set_actual_page_size(self.printable_entries(painter));
                     } else {
                         self.row_position = 0;
                         self.page = 0;
-                        self.update_values(line_buffer, completer);
+                        self.update_values(editor, completer);
                     }
                 }
                 MenuEvent::PreviousPage => {
@@ -563,7 +585,7 @@ impl Menu for ListMenu {
                         Some(page_num) => self.page = page_num,
                         None => self.page = self.pages.len().saturating_sub(1),
                     }
-                    self.update_values(line_buffer, completer);
+                    self.update_values(editor, completer);
                 }
             }
 
@@ -574,8 +596,18 @@ impl Menu for ListMenu {
     /// Calculates the real required lines for the menu considering how many lines
     /// wrap the terminal and if an entry is larger than the remaining lines
     fn menu_required_lines(&self, terminal_columns: u16) -> u16 {
-        self.get_values().iter().fold(0, |acc, suggestion| {
-            acc + self.number_of_lines(&suggestion.value, terminal_columns)
+        let mut entry_index = 0;
+        self.get_values().iter().fold(0, |total_lines, suggestion| {
+            //  to account for the the index and the indicator e.g. 0: XXXX
+            let ret = total_lines
+                + self.number_of_lines(
+                    &suggestion.value,
+                    terminal_columns.saturating_sub(
+                        self.indicator().width() as u16 + count_digits(entry_index),
+                    ),
+                );
+            entry_index += 1;
+            ret
         }) + 1
     }
 
@@ -655,6 +687,19 @@ fn number_of_lines(entry: &str, max_lines: usize, terminal_columns: u16) -> u16
     lines
 }
 
+fn count_digits(mut n: usize) -> u16 {
+    // count the digits in the number
+    if n == 0 {
+        return 1;
+    }
+    let mut count = 0;
+    while n > 0 {
+        n /= 10;
+        count += 1;
+    }
+    count
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
