@@ -3,10 +3,10 @@ use {
     crate::{
         menu::{Menu, ReedlineMenu},
         painting::PromptLines,
-        Prompt,
+        LineBuffer, Prompt,
     },
     crossterm::{
-        cursor::{self, MoveTo, RestorePosition, SavePosition},
+        cursor::{self, MoveTo, MoveUp, RestorePosition, SavePosition},
         style::{Attribute, Print, ResetColor, SetAttribute, SetForegroundColor},
         terminal::{self, Clear, ClearType, ScrollUp},
         QueueableCommand, Result,
@@ -81,13 +81,6 @@ impl Painter {
         self.screen_height() - self.prompt_start_row
     }
 
-    /// Check if the currently painted content exceeds the size of the screen
-    /// and thus should not be repainted without reason (disable animation
-    /// repaint)
-    pub(crate) fn exceeds_screen_size(&self) -> bool {
-        self.large_buffer
-    }
-
     /// Sets the prompt origin position and screen size for a new line editor
     /// invocation
     ///
@@ -95,7 +88,16 @@ impl Painter {
     /// [`Painter::handle_resize()`] instead
     pub(crate) fn initialize_prompt_position(&mut self) -> Result<()> {
         // Update the terminal size
-        self.terminal_size = terminal::size()?;
+        self.terminal_size = {
+            let size = terminal::size()?;
+            // if reported size is 0, 0 -
+            // use a default size to avoid divide by 0 panics
+            if size == (0, 0) {
+                (80, 24)
+            } else {
+                size
+            }
+        };
         // Cursor positions are 0 based here.
         let (column, row) = cursor::position()?;
         // Assumption: if the cursor is not on the zeroth column,
@@ -423,7 +425,17 @@ impl Painter {
         self.stdout.queue(MoveTo(0, 0))?;
         self.stdout.queue(cursor::Show)?;
 
-        self.stdout.flush()
+        self.stdout.flush()?;
+        self.initialize_prompt_position()
+    }
+
+    pub(crate) fn clear_scrollback(&mut self) -> Result<()> {
+        self.stdout
+            .queue(crossterm::terminal::Clear(ClearType::All))?
+            .queue(crossterm::terminal::Clear(ClearType::Purge))?
+            .queue(cursor::MoveTo(0, 0))?
+            .flush()?;
+        self.initialize_prompt_position()
     }
 
     // The prompt is moved to the end of the buffer after the event was handled
@@ -440,6 +452,51 @@ impl Painter {
 
         self.stdout.flush()
     }
+
+    /// Prints an external message
+    pub fn print_external_message(
+        &mut self,
+        messages: Vec<String>,
+        line_buffer: &LineBuffer,
+        prompt: &dyn Prompt,
+    ) -> Result<()> {
+        // adding 3 seems to be right for first line-wrap
+        let prompt_len = prompt.render_prompt_right().len() + 3;
+        let mut buffer_num_lines = 0_u16;
+        for (i, line) in line_buffer.get_buffer().lines().enumerate() {
+            let screen_lines = match i {
+                0 => {
+                    // the first line has to deal with the prompt
+                    let first_line_len = line.len() + prompt_len;
+                    // at least, it is one line
+                    ((first_line_len as u16) / (self.screen_width())) + 1
+                }
+                _ => {
+                    // the n-th line, no prompt, at least, it is one line
+                    ((line.len() as u16) / self.screen_width()) + 1
+                }
+            };
+            // count up screen-lines
+            buffer_num_lines = buffer_num_lines.saturating_add(screen_lines);
+        }
+        // move upward to start print if the line-buffer is more than one screen-line
+        if buffer_num_lines > 1 {
+            self.stdout.queue(MoveUp(buffer_num_lines - 1))?;
+        }
+        let erase_line = format!("\r{}\r", " ".repeat(self.screen_width().into()));
+        for line in messages {
+            self.stdout.queue(Print(&erase_line))?;
+            self.paint_line(&line)?;
+            let new_start = self.prompt_start_row.saturating_add(1);
+            let height = self.screen_height();
+            if new_start >= height {
+                self.prompt_start_row = height - 1;
+            } else {
+                self.prompt_start_row = new_start;
+            }
+        }
+        Ok(())
+    }
 }
 
 #[cfg(test)]
